//
// Public domain source by Philippe Elsass
// Last update: 2007/31/05
//
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#include <deque>
#include <string>

#include "localconnect.h"

/*************************************************** LOCAL CONNECTION */



// shared memory block size
static const int memSize = 65535;
// shared memory listeners offset
static const int listenersOffset = 40976;

// localconnection listener name (always localhost:<name>)
static const char listenerConnectionName[] = "localhost:fromExe";
static const char protocolName[] = "localhost";
static const char listenerMethod[] = "exeSays";

// localconnection emitter name (always localhost:<name>)
static const char emitterConnectionName[] = "localhost:toExe";
static const char emitterMethod[] = "swfSays";

typedef struct Local_Connection_Message_struct {
            std::string         in_message ;
            Local_Connection_Message_Type   type ;
        } Local_Connection_Message ;


typedef struct Local_Connection_State_struct {
            char *              buffer ;
            std::string         out_message ;
            std::string         in_message ;
            bool                listening , test_mode ;
            std::deque<Local_Connection_Message> queue ;
        } Local_Connection_State ;






static void messageQueue ( Local_Connection_State * lcs ,
                Local_Connection_Message_Type mt ,
                std::string & msg_text )
{
    Local_Connection_Message msg ;

    msg . in_message = msg_text ;
    msg . type = mt ;
    lcs -> queue . push_front ( msg ) ;
}

static void errorQueue ( Local_Connection_State * lcs ,
                Local_Connection_Message_Type mt )
{
    messageQueue ( lcs , mt , std::string ( "" ) ) ;
}



static int scanListeners ( Local_Connection_State * lcs ,
                const char * match )
{
	int i = listenersOffset;
	int alreadyListening = 0;
	while ( (i<memSize) && lcs->buffer[i] ) {
		if (!alreadyListening 
        && (strcmp(match, (char*)&lcs->buffer[i]) == 0)) {
			alreadyListening = 1;
			//printf("%4Xh: %s (myself)\n", i, (char*)&lcs->buffer[i]);
		}
		//else printf("%4Xh: %s\n", i, (char*)&lcs->buffer[i]);
		i += strlen((char*)&lcs->buffer[i])+1;
	}
	return (alreadyListening) ? -1 : i;
}

/**
 * Explore the listeners list
 * @return free offset to add listener, or -1 if already registered
 */
static int getFreeListenerIndex ( Local_Connection_State * lcs )
{
    return scanListeners ( lcs , emitterConnectionName ) ;
}


/**
 * Check if recipient is listening
 * @return free offset to add listener, or -1 if already registered
 */
static int checkEmitter ( Local_Connection_State * lcs )
{
    return scanListeners ( lcs , listenerConnectionName ) ;
}

/**
 * Add itself as a localconnection listener
 */
static void addlistener ( Local_Connection_State * lcs )
{
	int i = getFreeListenerIndex ( lcs );
	if (i>=memSize) {
		//printf("Registering failed\n");
        errorQueue ( lcs , LOCALCONN_ERROR_REGISTERING_FAILED ) ;
		Sleep(100);
		return;
	}
	lcs -> listening = true ;
	if (i > 0) {
		memcpy((char*)&lcs->buffer[i], 
                emitterConnectionName,
                strlen(emitterConnectionName) + 1);
		//printf("%4Xh: %s (myself)\n", i, emitterConnectionName);
	}
}

/**
 * Write an AMF (short) string in the buffer
 * @param str   text to write
 * @param pos   position in buffer
 * @return updated position in buffer
 */
static int writeAMFString ( Local_Connection_State * lcs ,
                const char * str , int pos )
{
	int len = strlen(str);
    len &= 0xff ;
	lcs->buffer[pos++] = 0x02;
	lcs->buffer[pos++] = 0; // string length is badly encoded here!
	lcs->buffer[pos++] = (char)len;
	memcpy((char*)&lcs->buffer[pos], str, len + 1);
	pos += len;
	return pos;
}

/**
 * Write message in memory file
 * @param msg   the text to send
 */
static int writeMessage ( Local_Connection_State * lcs , std::string msg )
{
	DWORD *timestamp = (DWORD*)&lcs->buffer[8];
	DWORD *size = (DWORD*)&lcs->buffer[12];
	int pos = 16;
	// check timestamp and size
	if (*size) return 0;
	*timestamp = GetTickCount();
	// write connection name
	pos = writeAMFString(lcs,listenerConnectionName, pos);
	// write protocol
	pos = writeAMFString(lcs,protocolName, pos);
	// write method name
	pos = writeAMFString(lcs,listenerMethod, pos);
	// write message
	pos = writeAMFString(lcs,msg.c_str(), pos);
	*size = pos-16;
	// report
	//printf("Written: timestamp=%d, size=%d\n", *timestamp, *size);
	// dump memory
	//dumpMemory(0, *size+16);
	return 1;
}

/**
 * Decode message in memory file: read (short) AMF string
 */
static void readMessage ( Local_Connection_State * lcs , int msgIndex )
{
	DWORD *timestamp = (DWORD*)&lcs->buffer[8];
	DWORD *size = (DWORD*)&lcs->buffer[12];
	int len;
	// check timestamp and size
	if (!(*size)) return;

	//printf("Received: timestamp=%d (current=%d), size=%d\n", *timestamp, GetTickCount(), *size);
	// dump memory
	//dumpMemory(0, *size+16);
	// read message
	len = lcs->buffer[msgIndex];
	if (len > 0 && len < MAX_SIZE)
	{
        char msg [ MAX_SIZE + 1 ] ;

		memcpy(msg, &lcs->buffer[msgIndex+1], len);
		msg[len] = 0;
        messageQueue ( lcs , LOCALCONN_MESSAGE , std::string ( msg ) ) ;
		//printf("Message: %s\n", msg);
	}
	// mark message as "read"
	lcs->buffer[8] = 0L;
	lcs->buffer[12] = 0L;
}

/**
 * check the memory file
 */
static int pollMemory ( Local_Connection_State * lcs )
{
	HANDLE hMutex;
	HANDLE hMapFile;
	LPVOID lpMapAddress;
	DWORD dwWaitResult;
	int result = 0;
	int len;
	int index;
	char dest[MAX_SIZE + 1 ];
	char method[MAX_SIZE + 1 ];

	// file access mutex
	hMutex = OpenMutex(
		MUTEX_ALL_ACCESS, // request full access
		FALSE, // handle not inheritable
		"MacromediaMutexOmega"); // object name

	// wait for mutex
	dwWaitResult = WaitForSingleObject(
		hMutex, // handle to mutex
		5000L); // five-second time-out interval

	switch (dwWaitResult) {
		// The thread got mutex ownership.
		case WAIT_OBJECT_0:
			try {
				// open file mapping
				hMapFile = OpenFileMapping(FILE_MAP_ALL_ACCESS, // read/write permission
					FALSE, // do not inherit the name
					"MacromediaFMOmega"); // mapping object name

				// get memory file pointer
				lpMapAddress = MapViewOfFile(hMapFile, // handle to mapping object
					FILE_MAP_ALL_ACCESS, // read/write permission
					0, // max. object size
					0, // size of hFile
					0); // map entire file

				// read memory
				if (lpMapAddress != NULL) {
					lcs->buffer = (char*)lpMapAddress;

					/* READING */

					// add itself to listeners
					if (!lcs->listening) 
                    {
                        addlistener(lcs);
                    }
					// if the recipient is correct, read the message
                    if ( lcs->buffer[16] ) 
                    {
						index = 16+2;
						len = lcs->buffer[index];
						memcpy(dest, &lcs->buffer[index+1], len);
						dest[len] = 0;
						if (strcmp(dest, emitterConnectionName) == 0)
						{
							index += 1+len+2;
							index += 1+(lcs->buffer[index])+2;
							// if the method is correct
							len = lcs->buffer[index];
							memcpy(method, &lcs->buffer[index+1], len);
							method[len] = 0;
							if (strcmp(method, emitterMethod) == 0)
							{
								// read message data
								readMessage(lcs, index+1+len+2);
								// handshake
								//lcs -> message = handshakeMessage;
							}
						}
					}

					/* SENDING */

					// check recipient
					if (checkEmitter ( lcs ) != -1) {
                        errorQueue ( lcs , 
                                LOCALCONN_ERROR_RECIPIENT_NOT_CONNECTED ) ;
						Sleep(100);
					}
					// no message waiting
					else if (( lcs -> out_message . size () != 0 )
                    && writeMessage ( lcs , lcs -> out_message )) {
                        lcs -> out_message = "" ;
                        messageQueue ( lcs , LOCALCONN_SENT_OK , 
                                std::string ( "" ) ) ;
						result = 0; // waiting for next message to send
					}
					//else printf("Message waiting\n");
				}
				else {
                    errorQueue ( lcs , 
                            LOCALCONN_ERROR_MAP_FAILED ) ;
					//printf("MapViewOfFile failed\n");
					result = -1;
				}

				// release file mapping
				if (!UnmapViewOfFile(lpMapAddress)) {
                    errorQueue ( lcs , 
                            LOCALCONN_ERROR_UNMAP_FAILED ) ;
					//printf("UnmapViewOfFile failed\n");
					result = -2;
				}
				if(!CloseHandle(hMapFile)) {
                    errorQueue ( lcs , 
                            LOCALCONN_ERROR_CLOSEHANDLE_FAILED ) ;
					// printf("CloseHandle failed\n");
					result = -3;
				}
			} catch ( ... ) {}

            // Release ownership of the mutex object.
            if(!ReleaseMutex(hMutex)) {
                //printf("ReleaseMutex failed\n");
                errorQueue ( lcs , 
                        LOCALCONN_ERROR_RELEASE_MUTEX_FAILED ) ;
                result = 1;
            }
            if(!CloseHandle(hMutex)) {
                //printf("CloseHandle failed\n");
                errorQueue ( lcs , 
                        LOCALCONN_ERROR_CLOSEHANDLE_FAILED ) ;
                result = 2;
            }
			break;

		// Cannot get mutex ownership due to time-out.
		case WAIT_TIMEOUT:
            errorQueue ( lcs , 
                    LOCALCONN_ERROR_WAIT_TIMEOUT ) ;
			//printf("Wait timeout\n");
			Sleep(100);
			break;
		// Got ownership of the abandoned mutex object.
		case WAIT_ABANDONED:
            errorQueue ( lcs , 
                    LOCALCONN_ERROR_WAIT_ABANDONED ) ;
			//printf("Wait abandoned\n");
			Sleep(100);
			break;
		default:
			//printf("No LocalConnection open\n");
			//printf(".");
			Sleep(100);
			break;
	}
	//
	Sleep(10);
	return result;
}



Local_Connection_State * LocalConn_New ( int test_mode )
{
    Local_Connection_State * lcs = new Local_Connection_State () ;

    lcs -> listening = false ;
    lcs -> test_mode = ( test_mode != 0 ) ;
    return lcs ;
}

Local_Connection_Message_Type LocalConn_Poll ( Local_Connection_State * lcs )
{
    if ( lcs -> test_mode )
    {
        if ( lcs -> out_message . size () != 0 )
        {
            messageQueue ( lcs , LOCALCONN_MESSAGE , lcs -> out_message ) ;
            lcs -> out_message = "" ;
        } else {
            errorQueue ( lcs , LOCALCONN_ERROR_TEST_ERROR ) ;
        }
    } else {
        pollMemory ( lcs ) ;
    }

    if ( lcs -> queue . size () == 0 )
    {
        return LOCALCONN_NOTHING ;
    }

    Local_Connection_Message lcm = lcs -> queue . back () ;
    Local_Connection_Message_Type type = lcm . type ;

    lcs -> queue . pop_back () ;

    if ( type == LOCALCONN_MESSAGE )
    {
        lcs -> in_message = lcm . in_message ;
    }
    return type ;
}

int LocalConn_Send ( Local_Connection_State * lcs , std::string msg_text )
{
    if ( lcs -> out_message . size () != 0 )
    {
        return 0 ;
    }
    lcs -> out_message = msg_text ;
    return 1 ;
}

std::string LocalConn_Receive ( Local_Connection_State * lcs )
{
    return lcs -> in_message ;
}

void LocalConn_Delete ( Local_Connection_State * lcs )
{
    delete lcs ;
}

const char * LocalConn_Get_Error ( Local_Connection_Message_Type mt )
{
    switch ( mt )
    {
		case LOCALCONN_ERROR_CLOSEHANDLE_FAILED :
            return "CloseHandle failed" ;
		case LOCALCONN_ERROR_MAP_FAILED :
            return "MapViewOfFile failed" ;
		case LOCALCONN_ERROR_RECIPIENT_NOT_CONNECTED :
            return "Recipient not connected" ;
		case LOCALCONN_ERROR_REGISTERING_FAILED :
            return "Registering failed" ;
		case LOCALCONN_ERROR_RELEASE_MUTEX_FAILED :
            return "ReleaseMutex failed" ;
		case LOCALCONN_ERROR_UNMAP_FAILED :
            return "UnmapViewOfFile failed" ;
		case LOCALCONN_ERROR_WAIT_ABANDONED :
			return "Wait abandoned" ;
		case LOCALCONN_ERROR_WAIT_TIMEOUT :
			return "Wait timeout" ;
		case LOCALCONN_ERROR_TEST_ERROR :
			return "Test error message" ;
		case LOCALCONN_MESSAGE :
		case LOCALCONN_NOTHING :
		case LOCALCONN_SENT_OK :
            return "" ;   /* not an error */
    }
	return "Unknown error" ;
}



